//
//  Triangulation.java
//  Triangulation
//
//  Created by Thomas Jannaud on 16/12/09.
//  Copyright (c) 2009 Jannaud's Technical Company. All rights reserved.
//

import java.awt.*;
import java.applet.*;
import javax.swing.event.*;
import javax.swing.*;
import java.awt.event.*;
import java.util.Collection;
import java.util.List;
import java.util.Iterator;
import java.util.HashSet;
import java.util.HashMap;
import java.lang.Math;

public class Triangulation extends JApplet implements Runnable, MouseListener, MouseMotionListener, ActionListener, ChangeListener {
	monInterfaceGraphique interfaceGUI;
	int lastX = -1, lastY = -1;
	int w = 0, h = 0;
	int offsetX = 10, offsetY = 10;
	nuagePoints points = new nuagePoints();
	monPoint pointCourant = null;
	boolean drawTriangulation = true;
	boolean drawVoronoi = false;
	
	public void run(){
		while (true) {
			repaint();
			try { Thread.sleep(2); }
			catch(InterruptedException e){}
		}
	}
	
	public void init() {
		interfaceGUI = new monInterfaceGraphique(this, getContentPane());
	}
	
	public void update(Graphics g) {
		paint(g);
	}
	
	private void dessinerPoint(monPoint p, Graphics g) {
		g.fillRect(fx(p.x) - 1, fy(p.y) - 1, 3, 3);
//		g.drawString(p.toString(), fx(p.x), fy(p.y));
	}
	
	private void dessinerLigne(monPoint p, monPoint q, Graphics g) {
		g.drawLine(fx(p.x), fy(p.y), fx(q.x), fy(q.y));
	}
	
	private void dessinerLignes(Collection<monPoint> lespoints, Graphics g) {
		monPoint pini = null;
		for(monPoint p : lespoints) {
			if (pini != null)
				g.drawLine(fx(pini.x), fy(pini.y), fx(p.x), fy(p.y));
			pini = p;
		}
	}
	
	public void dessinePolygone (monPoint[] polygone, Graphics g) {
		final int n = polygone.length;
		for (int i = 0; i < (n - 1); i++)
			g.drawLine(fx(polygone[i].x), fy(polygone[i].y), fx(polygone[i + 1].x), fy(polygone[i + 1].y));
	}
	
	private void dessinerTriangles(monTriangle s, Graphics g) {
		for(int i = 0; i < 3; i++)
			g.drawLine(fx(s.p[i].x),  fy(s.p[i].y), fx(s.p[(i + 1) % 3].x), fy(s.p[(i + 1) % 3].y));
	}
	
	public void paint (Graphics g) {
		super.paint(g);
		w = g.getClipBounds().width; h = interfaceGUI.pCommandes.getY();
		offsetX = Math.max(10 + (w - h) / 2, 10);
		offsetY = Math.max(10 + (h - w) / 2, 10);
		w -= 2 * offsetX;
		h -= 2 * offsetY;
		
		g.setColor(Color.blue);
		g.drawString("Thomas Jannaud 2009", 1, 10);
		for(monPoint p : points.points)
			dessinerPoint(p, g);
		
		g.setColor(Color.blue);
		if (drawTriangulation)
			for(monTriangle t : points.triangles.keySet())
				dessinerTriangles(t, g);
		
		g.setColor(Color.green);
		if (drawVoronoi)
			dessinerVoronoi(points.triangles, g);
	}
	
	private int fx(double x) {
		return offsetX + (int)(x * w);
	}
	
	private int fy(double y) {
		return offsetY + (int)((1. - y) * h);
	}
	
	public void mousePressed(MouseEvent event) {
		int x = event.getX();
		int y = event.getY();
		for(monPoint p : points.points) {
			int px = fx(p.x), py = fy(p.y);
			if (Math.abs(px - x) <= 2 && Math.abs(py - y) <= 2) {
				pointCourant = p;
				return;
			}
		}
		pointCourant = null;
		lastX = x;
		lastY = y;
	}
	
	public void mouseDragged(MouseEvent event) {
		if (pointCourant == null)
			return;
		// pointCourant.x = (double)(event.getX() - offsetX) / (double)(w);
		// pointCourant.y = 1. - (double)(event.getY() - offsetY) / (double)(h);
		repaint();
	}
	
	public void mouseReleased(MouseEvent event) {
		if (pointCourant == null && (Math.abs(lastX - event.getX()) + Math.abs(lastY - event.getY()) < 20)) {
			double x = (double)(event.getX() - offsetX) / (double)(w);
			double y =  1. - (double)(event.getY() - offsetY) / (double)(h);
			if (x > 0 && y > 0 && x < 1 && y < 1)
				points.ajouter(x, y);
			repaint();
		}
		pointCourant = null;
	}
	
	public void mouseClicked(MouseEvent event) {}
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}
	public void mouseMoved(MouseEvent e) {}
	
	public void actionPerformed(ActionEvent action) {
		if (action.getActionCommand().equals("Voronoi"))
			drawVoronoi = !drawVoronoi;
		else if (action.getActionCommand().equals("Triangulation"))
			drawTriangulation = !drawTriangulation;
		else if (action.getActionCommand().equals("Clear"))
			points.clear();
		else
			points.ajouter(interfaceGUI.slideNb.getValue());
		repaint();
	}
	
	public void stateChanged(ChangeEvent e) {
		JSlider source = (JSlider)e.getSource();
		int val = (int)source.getValue();
		interfaceGUI.boutonAdd.setText("add " + val + " random point" + (val == 1 ? "" : "s"));
	}
	
	private void dessinerVoronoi(HashMap<monTriangle, monTriangle> triangles, Graphics g) {
		if (points.points.size() < 3)
			return;
		
		HashSet<monPoint> done = new HashSet<monPoint>();
		for (monTriangle triangle : triangles.keySet()) {
//			dessinerPoint(triangle.getCentreCirconscrit(), g);
			for (int i = 0; i < 3; i++) {
				monPoint site = triangle.p[i];
				if (done.contains(site))
					continue;
				done.add(site);
				List<monTriangle> list = triangle.surroundingTriangles(site, points.segments);
				monPoint[] vertices = new monPoint[list.size()];
				int j = 0;
//				System.out.println("liste des triangles de " + site);
				for (monTriangle tri : list) {
//					System.out.println(j + " " + tri.petittoString());
					vertices[j++] = tri.getCentreCirconscrit();
				}
				dessinePolygone(vertices, g);
			}
		}
		
		// médiatrices fuyantes de l'enveloppe convexe
		int n = points.enveloppe_convexe.size();
		monPoint pini = points.enveloppe_convexe.get(n - 1);
		for(int i = 0; i < n; i++) {
			monPoint p = points.enveloppe_convexe.get(i);
//			System.out.println("dessin de mediatrice " + p + ", " + pini);
			monSegment seg = points.segments.get(new monSegment(p, pini));
			if (seg == null)
				System.out.println("ERREUR : seg ne devrait pas être nul : " + p + " et " + pini);
			monPoint q = (seg.p1 != null) ? seg.p1 : seg.p2;
			if (q == null)
				System.out.println("ERREUR : q ne devrait pas être nul : " + p + " et " + pini);
			monTriangle tr = new monTriangle(pini, p, q);
			dessinerSuperMediatrice(pini, p, tr, g);
			pini = p;
		}
	}
	
	private void dessinerSuperMediatrice(monPoint p, monPoint q, monTriangle tr, Graphics g) {
		monPoint centre = tr.getCentreCirconscrit(), milieu = new monPoint(.5 * (p.x + q.x), .5 * (p.y + q.y));
		monPoint a, b, start = centre;
		if (centre.isIn(tr)) {
			a = centre;
			b = milieu;
		}
		else if ((new monTriangle(p, q, centre)).isClockwise()) {
			a = centre;
			b = milieu;
		}
		else {
			a = milieu;
			b = centre;
		}
		// tracer une demi-droite partant de centre, direction a->b et allant taper le cadre
		double vx = b.x - a.x, vy = b.y - a.y;
		double choixX = (vx > 0) ? 1. : 0.;
		double choixY = (vy > 0) ? 1. : 0.;
		double det = Math.abs((choixX - start.x) * vy) - Math.abs((choixY - start.y) * vx);		
		double lambda = -1.;
		if (det > 0) {
			lambda = (choixY - start.y) / vy;
			choixX = start.x + lambda * vx;
		}
		else {
			lambda = (choixX - start.x) / vx;
			choixY = start.y + lambda * vy;
		}
		if (lambda >= 0)
			dessinerLigne(start, new monPoint(choixX, choixY), g);
		// System.out.println("super med : vx:" + vx + ", vy:" + vy + ", choixX:" + choixX + ", choixY:" + choixY + ", det:" + det + ", l:" + lambda);
	}
}