//
//  monTriangle.java
//  Triangulation
//
//  Created by Thomas Jannaud on 16/12/09.
//  Copyright 2009 Jannaud's Technical Company. All rights reserved.
//

import java.util.List;
import java.util.LinkedList;
import java.util.HashMap;

public class monTriangle {
	public monPoint[] p;
	public monSegment[] segments;
	public monTriangle(monPoint pt1, monPoint pt2, monPoint pt3) {
		p = new monPoint[3];
		p[0] = pt1;
		p[1] = pt2;
		p[2] = pt3;
		segments = null;
	}
	public monTriangle(monPoint pt1, monPoint pt2, monPoint pt3, monSegment s12, monSegment s23, monSegment s31) {
		p = new monPoint[3];
		p[0] = pt1;
		p[1] = pt2;
		p[2] = pt3;
		segments = new monSegment[3];
		segments[0] = s12;
		segments[1] = s23;
		segments[2] = s31;
	}
	public static boolean isClockwise(monPoint pt1, monPoint pt2, monPoint pt3) {
		monVecteur v12 = new monVecteur(pt1, pt2);
		monVecteur v23 = new monVecteur(pt1, pt3);
		return (v12.determinant(v23) < 0.);
	}
	public boolean isClockwise() {
		monVecteur v12 = new monVecteur(p[0], p[1]);
		monVecteur v23 = new monVecteur(p[0], p[2]);
		return (v12.determinant(v23) < 0.);
	}
	public boolean estSommet(monPoint q) {
		for(int i = 0; i < 3; i++)
			if (q.equals(p[i]))
				return true;
		return false;
	}
	public int hashCode() {
		return p[0].hashCode() + p[1].hashCode() + p[2].hashCode();
	}
	public boolean equals(Object obj) {
		if (obj instanceof monTriangle) {
			final monTriangle t = (monTriangle)(obj);
			final int[][] permutations = {{0, 1, 2}, {0, 2, 1}, {1, 0, 2}, {1, 2, 0}, {2, 0, 1}, {2, 1, 0}};
			for(int i = 0; i < 6; i++) {
				if (p[0].equals(t.p[permutations[i][0]]) && p[1].equals(t.p[permutations[i][1]]) && p[2].equals(t.p[permutations[i][2]]))
					return true;
			}
			return false;
		}
		else
			return false;
	}
	public void replace(monPoint a, monPoint b) {
		// les segments sont mis à jour indépendamment
		for(int i = 0; i < 3; i++) {
			if (p[i].equals(a)) {
				p[i] = b;
				return;
			}
		}
		System.out.println("ERREUR !!!!! replace " + a + " par " + b + "\ndans triangle :" + this);
	}
	public void replace(monSegment a, monSegment b) {
		for(int i = 0; i < 3; i++) {
			if (segments[i].equals(a)) {
				segments[i] = b;
				return;
			}
		}
		System.out.println("ERREUR !!!!! replace " + a + " par " + b + "\ndans triangle :" + this);
	}
	public String toString() {
		return "Triangle : p0:" + p[0] + ", p1:" + p[1] + ", p2:" + p[2] + " **** segments s0:" + segments[0] + ", s1:" + segments[1] + ", s2:" + segments[2];
	}
	public String petittoString() {
		return "Triangle (" + p[0] + ", " + p[1] + ", " + p[2] + ")";
	}
	public boolean isIn(monTriangle t) {
		for(int i = 0; i < 3; i++)
			if (!p[i].isIn(t))
				return false;
		return true;
	}
	public boolean hasEdge(monSegment s) {
		//	System.out.println("has edge : " + s.toString() + " about " + petittoString());
		int n = 0;
		for(int i = 0; i < 3; i++) {
			if (s.A.equals(p[i]) || s.B.equals(p[i]))
				n++;
		}
		//		System.out.println("n = " + n);
		return (n == 2);
	}
	public List<monTriangle> surroundingTriangles(monPoint q, HashMap<monSegment, monSegment> allSegments) {
		if (!estSommet(q)) {
			System.out.println("ERREUR : surrounding triangle " + this + ", n'a pas pour sommet q =" + q);
			return null;
		}
		//		System.out.println("demande surrounding de " + q + " pour triangle " + toString());
		LinkedList<monTriangle> ret = new LinkedList<monTriangle>();
		ret.add(this);
		boolean first = true;
		for(int i = 0; i < 3; i++) {
			if (!segments[i].aPourSommet(q))
				continue;
			monSegment seg = segments[i];
			monPoint lastP = p[(i + 2) % 3];
			while (true) {
				//				System.out.println("entrée : seg=" + seg + ", lastP : " + lastP);
				monPoint nouvP = null;
				if (seg.p1 != null && !seg.p1.equals(lastP))
					nouvP = seg.p1;
				else if (seg.p2 != null && !seg.p2.equals(lastP))
					nouvP = seg.p2;
				if (nouvP == null)
					break;
				monTriangle tr = new monTriangle(seg.A, seg.B, nouvP);
				if (this.equals(tr)) {
					if (!first)
						System.out.println("ERREUR : si on fait le tour du triangle, on devrait l'avoir fait dans l'autre sens avant ! " + tr);
					ret.addLast(tr);
					return ret;
				}
				//				System.out.println("surrounding ajoute " + tr.petittoString());
				lastP = q.equals(seg.A) ? seg.B : seg.A;
				if (first)
					ret.addFirst(tr);
				else
					ret.addLast(tr);
				seg = allSegments.get(new monSegment(q, nouvP));
				if (seg == null)
					System.out.println("ERREUR : pas de segment par " + q.toString() + " et " + nouvP.toString() + " dans surroundingTriangles");
			}
			first = false;
			//			System.out.println("will add last");
		}
		return ret;
	}
	
	// centre du cercle circonscrit
	public monPoint getCentreCirconscrit() {
		double[][] matrix = new double[2][];
		for (int i = 0; i < 2; i++)
			matrix[i] = eqbizarre(p[i], p[i + 1]);
		double[] hCenter = prod_vectoriel(matrix);
		return new monPoint(hCenter[0] / hCenter[2], hCenter[1] / hCenter[2]);
	}
	
	private static double[] eqbizarre(monPoint q, monPoint r) {
		double a = r.x - q.x, b = r.y - q.y;
		double c = -.5 * (r.x * r.x + r.y * r.y - q.x * q.x - q.y * q.y);
		double [] ret = new double[3];
		ret[0] = a;
		ret[1] = b;
		ret[2] = c;
		return ret;
	}
	
	private static double[] prod_vectoriel(double[][] matrix) {
		double[] ret = new double[3];
		ret[0] = matrix[0][1] * matrix[1][2] - matrix[0][2] * matrix[1][1];
		ret[1] = matrix[0][2] * matrix[1][0] - matrix[0][0] * matrix[1][2];
		ret[2] = matrix[0][0] * matrix[1][1] - matrix[0][1] * matrix[1][0];
		return ret;
	}	
}